/********************************************************************\

  Name:         cd_hv.c
  Created by:   Stefan Ritt

  Contents:     High Voltage Class Driver


  Revision history
  -------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  26-MAR-97    SR    created


\********************************************************************/

#include <stdio.h>
#include "midas.h"
#include "msystem.h"

/* ODB keys */
static HNDLE  hDB, hKeyRoot;
static HNDLE  hKeyDemand, hKeyMeasured, hKeyCurrent;

/* globals */
static INT    num_channels;
static EQUIPMENT *pequipment;

/* items in /Variables record */
static char   *names;
static float  *demand;
static float  *measured;
static float  *current;

/* items in /Settings */
static float  *update_threshold;
static float  *update_threshold_current;
static float  *voltage_limit;
static float  *current_limit;
static float  *ramp_speed;

/* mirror arrays */
static float  *demand_mirror;
static float  *measured_mirror;
static float  *current_mirror;
static DWORD  *last_change;

static void   **driver;
static INT    *channel_offset;
static INT    *device_index;

#define DRIVER(_i) ((INT (*)(INT cmd, ...))(driver[_i]))

/*------------------------------------------------------------------*/

static void free_mem()
{
  free(names);
  free(demand);
  free(measured);
  free(current);

  free(update_threshold);
  free(update_threshold_current);
  free(voltage_limit);
  free(current_limit);
  free(ramp_speed);

  free(demand_mirror);
  free(measured_mirror);
  free(current_mirror);
  free(last_change);

  free(device_index);
  free(channel_offset);
  free(driver);
}

/*------------------------------------------------------------------*/

INT hv_read(int channel)
{
int          i, j, status;
float        max_diff;
DWORD        act_time, min_time;
BOOL         changed;
static DWORD last_update;

  if (channel == -1)
    {
    for (i=0,j=0 ; pequipment->driver[i].name[0] ; i++)
      {
      status = pequipment->driver[i].func(CMD_GET_ALL, 
                                          pequipment->driver[i].index, 
                                          pequipment->driver[i].channels, 
                                          measured+j);
      status = pequipment->driver[i].func(CMD_GET_CURRENT_ALL,
                                          pequipment->driver[i].index, 
                                          pequipment->driver[i].channels, 
                                          current+j);

      j += pequipment->driver[i].channels;
      }
    }
  else
    {
    status = DRIVER(channel)(CMD_GET, device_index[channel], channel-channel_offset[channel], &measured[channel]);
    status = DRIVER(channel)(CMD_GET_CURRENT, device_index[channel], channel-channel_offset[channel], &current[channel]);
    }

  /* check how much channels have changed since last ODB update */
  act_time = ss_millitime();

  /* check for update measured */
  max_diff = 0.f;
  min_time = 10000;
  changed  = FALSE;
  for (i=0 ; i<num_channels ; i++)
    {
    if ( abs(measured[i] - measured_mirror[i]) > max_diff)
      max_diff = abs(measured[i] - measured_mirror[i]);

    if ( abs(measured[i] - measured_mirror[i]) > update_threshold[i])
      changed = TRUE;

    if ( act_time - last_change[i] < min_time)
      min_time = act_time - last_change[i];
    }

  /* update if change is more than update_sensitivity or less than 2sec ago 
     or last update is older than a minute */
  if (changed || (min_time < 2000 && max_diff > 0) ||
      act_time - last_update > 60000)
    {
    last_update = act_time;

    for (i=0 ; i<num_channels ; i++)
      measured_mirror[i] = measured[i];

    db_set_data(hDB, hKeyMeasured, measured, sizeof(float)*num_channels, num_channels, TID_FLOAT);

    pequipment->odb_out++;
    }

  /* check for update current */
  max_diff = 0.f;
  min_time = 10000;
  changed  = FALSE;
  for (i=0 ; i<num_channels ; i++)
    {
    if ( abs(current[i] - current_mirror[i]) > max_diff)
      max_diff = abs(current[i] - current_mirror[i]);

    if ( abs(current[i] - current_mirror[i]) > update_threshold_current[i])
      changed = TRUE;

    if ( act_time - last_change[i] < min_time)
      min_time = act_time - last_change[i];
    }

  /* update if change is more than update_sensitivity or less than 2sec ago */
  if (changed || (min_time < 2000 && max_diff > 0))
    {
    for (i=0 ; i<num_channels ; i++)
      current_mirror[i] = current[i];

    db_set_data(hDB, hKeyCurrent, current, sizeof(float)*num_channels, num_channels, TID_FLOAT);

    pequipment->odb_out++;
    }

  return status;
}

/*------------------------------------------------------------------*/

INT hv_ramp()
{
INT i, status;
float delta;

  for (i=0 ; i<num_channels ; i++)
    if (demand[i] != demand_mirror[i])
      {
      if (ramp_speed[i] == 0)
        delta = demand[i] - demand_mirror[i];
      else
        delta = (float) ((ss_millitime() - last_change[i])/1000.0 * ramp_speed[i]);
      delta = (float) (INT) abs(delta);
      
      if (delta > 0)
        {
        if (demand[i] > demand_mirror[i])
          demand_mirror[i] = min(demand[i], demand_mirror[i] + abs(delta));
        else
          demand_mirror[i] = max(demand[i], demand_mirror[i] - abs(delta));

        status = DRIVER(i)(CMD_SET, device_index[i], i-channel_offset[i], demand_mirror[i]);
        last_change[i] = ss_millitime();
        }
      }

  return status;
}


/*------------------------------------------------------------------*/

void hv_demand(INT hDB, INT hKey)
{
INT   i, status;

  /* check for voltage limit */
  for (i=0 ; i<num_channels ; i++)
    if (demand[i] > voltage_limit[i])
      demand[i] = voltage_limit[i];

  /* check if all channels have same value and no ramping */
  for (i=1 ; i<num_channels ; i++)
    if (demand[i] != demand[0] || ramp_speed[i] != 0)
      break;

  if (i == num_channels)
    {
    /* set all channels to the same value */
    for (i=0 ; pequipment->driver[i].name[0] ; i++)
      status = pequipment->driver[i].func(CMD_SET_ALL, device_index[i], pequipment->driver[i].channels, 
                                          demand[0]);
    for (i=0 ; i<num_channels ; i++)
      {
      if (demand[i] != demand_mirror[i])
        last_change[i] = ss_millitime();
      demand_mirror[i] = demand[i];
      }

    /* read back all channels */
    hv_read(-1);
    }
  else
    /* set individual channels only if demand value differs */
    for (i=0 ; i<num_channels ; i++)
      if (demand[i] != demand_mirror[i] ||
          abs(measured[i] - demand[i]) > 100)
        {
        /* set directly if no ramping */
        if (ramp_speed[i] == 0)
          {
          status = DRIVER(i)(CMD_SET, device_index[i], i-channel_offset[i], demand[i]);
          demand_mirror[i] = demand[i];
          }

        last_change[i] = ss_millitime();
        }

  pequipment->odb_in++;

  /* execute one ramping */
  hv_ramp();
}

/*------------------------------------------------------------------*/

void hv_current_limit(INT hDB, INT hKey)
{
INT   i;

  /* check for voltage limit */
  for (i=0 ; i<num_channels ; i++)
    DRIVER(i)(CMD_SET_CURRENT_LIMIT, device_index[i], i-channel_offset[i], current_limit[i]);

  pequipment->odb_in++;
}

/*------------------------------------------------------------------*/

INT hv_init()
{
int   status, size, i, j, index, offset;
char  str[256];
HNDLE hKey;

  /* get class driver root key */
  cm_get_experiment_database(&hDB, NULL);
  sprintf(str, "/Equipment/%s", pequipment->name);
  db_create_key(hDB, 0, str, TID_KEY);
  db_find_key(hDB, 0, str, &hKeyRoot);

  /* count total number of channels */
  db_create_key(hDB, hKeyRoot, "Settings/Channels", TID_KEY);
  db_find_key(hDB, hKeyRoot, "Settings/Channels", &hKey);

  for (i=0,num_channels=0 ; pequipment->driver[i].name[0] ; i++)
    {
    /* ODB value has priority over driver list in channel number */
    size = sizeof(INT);
    db_get_value(hDB, hKey, pequipment->driver[i].name, 
                 &pequipment->driver[i].channels, &size, TID_INT);
    num_channels += pequipment->driver[i].channels;

    /* evaluate index */
    pequipment->driver[i].index = 0;
    
    for (j=0 ; j<i ; j++)
      if (pequipment->driver[j].func == pequipment->driver[i].func)
        pequipment->driver[i].index++;
    }

  if (num_channels == 0)
    {
    cm_msg(MERROR, "hv_init", "No channels found in device driver list");
    return FE_ERR_ODB;
    }

  /* Allocate memory for buffers */
  names            = (char *)  calloc(num_channels, NAME_LENGTH);

  demand           = (float *) calloc(num_channels, sizeof(float));
  measured         = (float *) calloc(num_channels, sizeof(float));
  current          = (float *) calloc(num_channels, sizeof(float));
  
  update_threshold = (float *) calloc(num_channels, sizeof(float));
  update_threshold_current = (float *) calloc(num_channels, sizeof(float));
  voltage_limit    = (float *) calloc(num_channels, sizeof(float));
  current_limit    = (float *) calloc(num_channels, sizeof(float));
  ramp_speed       = (float *) calloc(num_channels, sizeof(float));

  demand_mirror    = (float *) calloc(num_channels, sizeof(float));
  measured_mirror  = (float *) calloc(num_channels, sizeof(float));
  current_mirror   = (float *) calloc(num_channels, sizeof(float));
  last_change      = (DWORD *) calloc(num_channels, sizeof(DWORD));

  device_index     = (INT *)   calloc(num_channels, sizeof(INT));
  channel_offset   = (INT *)   calloc(num_channels, sizeof(INT));
  driver           = (void *)  calloc(num_channels, sizeof(void*));

  if (!driver)
    {
    cm_msg(MERROR, "hv_init", "Not enough memory");
    return FE_ERR_ODB;
    }

  /*---- Create/Read settings ----*/

  /* Names */
  for (i=0 ; i<num_channels ; i++)
    sprintf(names+NAME_LENGTH*i, "Default%%CH %d", i);
  db_merge_data(hDB, hKeyRoot, "Settings/Names", 
                names, NAME_LENGTH*num_channels, num_channels, TID_STRING);

  /* Update threshold */
  for (i=0 ; i<num_channels ; i++)
    update_threshold[i] = 2.f; /* default 2V */
  db_merge_data(hDB, hKeyRoot, "Settings/Update Threshold Measured", 
                update_threshold, sizeof(float) * num_channels, num_channels, TID_FLOAT);

  /* Update threshold current */
  for (i=0 ; i<num_channels ; i++)
    update_threshold_current[i] = 2.f; /* default 2uA */
  db_merge_data(hDB, hKeyRoot, "Settings/Update Threshold Current", 
                update_threshold_current, sizeof(float) * num_channels, num_channels, TID_FLOAT);

  /* Voltage limit */
  for (i=0 ; i<num_channels ; i++)
    voltage_limit[i] = 3000.f; /* default 3000V */
  db_merge_data(hDB, hKeyRoot, "Settings/Voltage Limit", 
                voltage_limit, sizeof(float) * num_channels, num_channels, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Voltage Limit", &hKey);
  db_open_record(hDB, hKey, voltage_limit, num_channels*sizeof(float), MODE_READ, NULL);

  /* Current limit */
  for (i=0 ; i<num_channels ; i++)
    current_limit[i] = 3000.f; /* default 3000V */
  db_merge_data(hDB, hKeyRoot, "Settings/Current Limit", 
                current_limit, sizeof(float) * num_channels, num_channels, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Current Limit", &hKey);
  db_open_record(hDB, hKey, current_limit, num_channels*sizeof(float), MODE_READ, hv_current_limit);

  /* Ramp speed */
  for (i=0 ; i<num_channels ; i++)
    ramp_speed[i] = 0.f; /* disabled by default */
  db_merge_data(hDB, hKeyRoot, "Settings/Ramping Speed", 
                ramp_speed, sizeof(float) * num_channels, num_channels, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Ramping Speed", &hKey);
  db_open_record(hDB, hKey, ramp_speed, num_channels*sizeof(float), MODE_READ, NULL);

  /*---- Create/Read variables ----*/

  /* Demand */
  db_merge_data(hDB, hKeyRoot, "Variables/Demand", 
                demand, sizeof(float)*num_channels, num_channels, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Variables/Demand", &hKeyDemand);
  db_open_record(hDB, hKeyDemand, demand, num_channels*sizeof(float), MODE_READ, hv_demand);

  /* Measured */
  db_merge_data(hDB, hKeyRoot, "Variables/Measured", 
                measured, sizeof(float)*num_channels, num_channels, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Variables/Measured", &hKeyMeasured);
  memcpy(measured_mirror, measured, num_channels * sizeof(float));

  /* Current */
  db_merge_data(hDB, hKeyRoot, "Variables/Current", 
                current, sizeof(float)*num_channels, num_channels, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Variables/Current", &hKeyCurrent);
  memcpy(current_mirror, current, num_channels * sizeof(float));

  /*---- Initialize device drivers ----*/

  /* call init method */
  for (i=0 ; pequipment->driver[i].name[0] ; i++)
    {
    sprintf(str, "Settings/Devices/%s", pequipment->driver[i].name);
    status = db_find_key(hDB, hKeyRoot, str, &hKey);
    if (status != DB_SUCCESS)
      {
      db_create_key(hDB, hKeyRoot, str, TID_KEY);
      status = db_find_key(hDB, hKeyRoot, str, &hKey);
      if (status != DB_SUCCESS)
        {
        cm_msg(MERROR, "hv_init", "Cannot create %s entry in online database", str);
        free_mem();
        return FE_ERR_ODB;
        }
      }

    status = pequipment->driver[i].func(CMD_INIT, hKey, pequipment->driver[i].index,
                                        pequipment->driver[i].channels);
    if (status != FE_SUCCESS)
      {
      free_mem();
      return status;
      }
    }

  /* compose device driver channel assignment */
  for (i=0,j=0,index=0,offset=0 ; i<num_channels ; i++,j++)
    {
    while (j >= pequipment->driver[index].channels && 
           pequipment->driver[index].name[0])
      {
      offset += j;
      index++;
      j = 0;
      }

    driver[i] = pequipment->driver[index].func;
    device_index[i] = pequipment->driver[index].index;
    channel_offset[i] = offset;
    }

  /* set limits and initial demand values */
  for (i=0 ; i<num_channels ; i++)
    {
    DRIVER(i)(CMD_SET_CURRENT_LIMIT, device_index[i], i-channel_offset[i], current_limit[i]);
    DRIVER(i)(CMD_SET, device_index[i], i-channel_offset[i], min(demand[i], voltage_limit[i]));
    }

  /* initially read all channels */
  hv_read(-1);

  return FE_SUCCESS;
}

/*------------------------------------------------------------------*/

INT hv_exit()
{
INT i;

  free_mem();

  /* call exit method of device drivers */
  for (i=0 ; pequipment->driver[i].func != NULL ; i++)
    pequipment->driver[i].func(CMD_EXIT, pequipment->driver[i].index);

  return FE_SUCCESS;
}

/*------------------------------------------------------------------*/

INT hv_idle()
{
INT        act, status;
static INT last=0;
DWORD      act_time;

  /* do ramping */
  hv_ramp();

  /* select next measurement channel */
  act_time = ss_millitime();
  act = (last + 1) % num_channels;

  /* look for the next channel recently updated */
  while (!( act_time - last_change[act] < 5000 ||
          ( act_time - last_change[act] < 20000 
          &&  abs(measured[act] - demand[act]) > 2*update_threshold[act])))
    {
    act = (act + 1) % num_channels;
    if (act == last)
      {
      act = (last + 1) % num_channels;
      break;
      }
    }

  /* measure channel */
  status = hv_read(act);
  last = act;

  return status;
}

/*------------------------------------------------------------------*/

INT cd_hv_read(char *pevent)
{
  memcpy(pevent, demand, sizeof(float)*num_channels);
  pevent += sizeof(float)*num_channels;
  
  memcpy(pevent, measured, sizeof(float)*num_channels);
  pevent += sizeof(float)*num_channels;

  memcpy(pevent, current, sizeof(float)*num_channels);
  pevent += sizeof(float)*num_channels;

  return 3*sizeof(float)*num_channels;
}

/*------------------------------------------------------------------*/

INT cd_hv(INT cmd, PTYPE adr)
{
INT  status;

  switch (cmd)
    {
    case CMD_INIT:
      pequipment = (EQUIPMENT *) adr;
      status = hv_init();
      break;

    case CMD_EXIT:
      status = hv_exit();
      break;

    case CMD_IDLE:
      status = hv_idle();
      break;

    default:
      cm_msg(MERROR, "HV class driver", "Received unknown command %d", cmd);
      status = FE_ERR_DRIVER;
      break;

    }

  return status; 
}
