/********************************************************************\

  Name:         cd_multi.c
  Created by:   Stefan Ritt

  Contents:     Multimeter Class Driver


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  02-APR-97    SR    created


\********************************************************************/

#include <stdio.h>
#include "midas.h"
#include "msystem.h"

/* ODB keys */
static HNDLE  hDB, hKeyRoot;
static HNDLE  hKeyOutput, hKeyInput;

/* globals */
static INT    num_channels_input, num_channels_output;
static EQUIPMENT *pequipment;

/* items in /Variables record */
static char   *names_input, *names_output;
static float  *var_input;
static float  *var_output;

/* items in /Settings */
static float  *update_threshold;
static float  *factor_input, *factor_output;
static float  *offset_input, *offset_output;

/* mirror arrays */
static float  *input_mirror;
static float  *output_mirror;

static void   **driver_input, **driver_output;
static INT    *channel_offset_input, *channel_offset_output;
static INT    *device_index_input, *device_index_output;

#define DRIVER_INPUT(_i) ((INT (*)(INT cmd, ...))(driver_input[_i]))
#define DRIVER_OUTPUT(_i) ((INT (*)(INT cmd, ...))(driver_output[_i]))

/*----------------------------------------------------------------------------*/

static void free_mem()
{
  free(names_input);
  free(names_output);

  free(var_input);
  free(var_output);
  
  free(update_threshold);
  free(offset_input);
  free(factor_input);
  free(offset_output);
  free(factor_output);

  free(input_mirror);
  free(output_mirror);

  free(device_index_input);
  free(device_index_output);
  free(channel_offset_input);
  free(channel_offset_output);

  free(driver_input);
  free(driver_output);
}

/*----------------------------------------------------------------------------*/

void multi_read(int channel)
{
int          i, status;
DWORD        actual_time;
static DWORD last_update;

  if (channel == -1)
    for (i=0 ; i<num_channels_input ; i++)
      {
      status = DRIVER_INPUT(i)(CMD_GET, device_index_input[i], i-channel_offset_input[i], &var_input[i]);
      var_input[i] = var_input[i] * factor_input[i] - offset_input[i];
      }
  else
    {
    status = DRIVER_INPUT(channel)(CMD_GET, device_index_input[channel], channel-channel_offset_input[channel], 
                          &var_input[channel]);
    var_input[channel] = var_input[channel] * factor_input[channel] - offset_input[channel];
    }

  /* check if significant change since last ODB update */
  for (i=0 ; i<num_channels_input ; i++)
    if ( abs(var_input[i] - input_mirror[i]) > update_threshold[i])
      break;

  /* update if change is more than update_sensitivity or last update more
     than a minute ago */
  actual_time = ss_millitime();
  if (i < num_channels_input || actual_time - last_update > 60000)
    {
    last_update = actual_time;

    for (i=0 ; i<num_channels_input ; i++)
      input_mirror[i] = var_input[i];

    db_set_data(hDB, hKeyInput, var_input, num_channels_input*sizeof(float), 
                num_channels_input, TID_FLOAT);

    pequipment->odb_out++;
    }
}

/*----------------------------------------------------------------------------*/

void multi_output(INT hDB, INT hKey)
{
INT   i;
DWORD act_time;

  act_time = ss_millitime();

  for (i=0 ; i<num_channels_output ; i++)
    {
    /* only set channel if demand value differs */
    if (var_output[i] != output_mirror[i])
      {
      DRIVER_OUTPUT(i)(CMD_SET, device_index_output[i], i-channel_offset_output[i], var_output[i]);

      output_mirror[i] = var_output[i];
      }
    }

  pequipment->odb_in++;
}

/*----------------------------------------------------------------------------*/

INT multi_init()
{
int   status, size, i, j, index, ch_offset;
char  str[256];
HNDLE hKey;

  /* get class driver root key */
  cm_get_experiment_database(&hDB, NULL);
  sprintf(str, "/Equipment/%s", pequipment->name);
  db_create_key(hDB, 0, str, TID_KEY);
  db_find_key(hDB, 0, str, &hKeyRoot);

  /* count total number of channels */
  db_create_key(hDB, hKeyRoot, "Settings/Channels", TID_KEY);
  db_find_key(hDB, hKeyRoot, "Settings/Channels", &hKey);

  for (i=num_channels_input=num_channels_output=0 ; pequipment->driver[i].name[0] ; i++)
    {
    /* ODB value has priority over driver list */
    size = sizeof(INT);
    db_get_value(hDB, hKey, pequipment->driver[i].name, 
                 &pequipment->driver[i].channels, &size, TID_INT);

    if (pequipment->driver[i].type == CH_INPUT)
      num_channels_input += pequipment->driver[i].channels;
    if (pequipment->driver[i].type == CH_OUTPUT)
      num_channels_output += pequipment->driver[i].channels;
    }

  if (num_channels_input == 0 && num_channels_output == 0)
    {
    cm_msg(MERROR, "multi_init", "No channels found in device driver list");
    return FE_ERR_ODB;
    }

  /* Allocate memory for buffers */
  names_input      = (char *)  calloc(num_channels_input, NAME_LENGTH);
  names_output     = (char *)  calloc(num_channels_output, NAME_LENGTH);

  var_input        = (float *) calloc(num_channels_input, sizeof(float));
  var_output       = (float *) calloc(num_channels_output, sizeof(float));
  
  update_threshold = (float *) calloc(num_channels_input, sizeof(float));
  offset_input     = (float *) calloc(num_channels_input, sizeof(float));
  factor_input     = (float *) calloc(num_channels_input, sizeof(float));
  offset_output    = (float *) calloc(num_channels_output, sizeof(float));
  factor_output    = (float *) calloc(num_channels_output, sizeof(float));

  input_mirror     = (float *) calloc(num_channels_input, sizeof(float));
  output_mirror    = (float *) calloc(num_channels_output, sizeof(float));

  device_index_input    = (INT *)   calloc(num_channels_input, sizeof(INT));
  device_index_output   = (INT *)   calloc(num_channels_output, sizeof(INT));
  channel_offset_input  = (INT *)   calloc(num_channels_input, sizeof(INT));
  channel_offset_output = (DWORD *) calloc(num_channels_output, sizeof(DWORD));

  driver_input     = (void *)  calloc(num_channels_input, sizeof(void*));
  driver_output    = (void *)  calloc(num_channels_output, sizeof(void*));

  if (!driver_output)
    {
    cm_msg(MERROR, "multi_init", "Not enough memory");
    return FE_ERR_ODB;
    }

  /*---- Create/Read settings ----*/

  /* Names */
  for (i=0 ; i<num_channels_input ; i++)
    sprintf(names_input+NAME_LENGTH*i, "Input Channel %d", i);
  db_merge_data(hDB, hKeyRoot, "Settings/Names Input", 
                names_input, num_channels_input*NAME_LENGTH, num_channels_input, TID_STRING);
  for (i=0 ; i<num_channels_output ; i++)
    sprintf(names_output+NAME_LENGTH*i, "Output Channel %d", i);
  db_merge_data(hDB, hKeyRoot, "Settings/Names Output", 
                names_output, num_channels_output*NAME_LENGTH, num_channels_output, TID_STRING);

  /* Update threshold */
  for (i=0 ; i<num_channels_input ; i++)
    update_threshold[i] = 0.1f; /* default 0.1 */
  db_merge_data(hDB, hKeyRoot, "Settings/Update Threshold", 
                update_threshold, num_channels_input*sizeof(float), num_channels_input, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Update Threshold", &hKey);
  db_open_record(hDB, hKey, update_threshold, num_channels_input*sizeof(float), MODE_READ, NULL);

  /* Offset */
  for (i=0 ; i<num_channels_input ; i++)
    offset_input[i] = 0.f; /* default 0 */
  db_merge_data(hDB, hKeyRoot, "Settings/Input Offset", 
                offset_input, num_channels_input*sizeof(float), num_channels_input, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Input Offset", &hKey);
  db_open_record(hDB, hKey, offset_input, num_channels_input*sizeof(float), MODE_READ, NULL);

  for (i=0 ; i<num_channels_output ; i++)
    offset_output[i] = 0.f;
  db_merge_data(hDB, hKeyRoot, "Settings/Output Offset", 
                offset_output, num_channels_output*sizeof(float), num_channels_output, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Output Offset", &hKey);
  db_open_record(hDB, hKey, offset_output, num_channels_output*sizeof(float), MODE_READ, NULL);

  /* Factor */
  for (i=0 ; i<num_channels_input ; i++)
    factor_input[i] = 1.f; /* default 1 */
  db_merge_data(hDB, hKeyRoot, "Settings/Input Factor", 
                factor_input, num_channels_input*sizeof(float), num_channels_input, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Input factor", &hKey);
  db_open_record(hDB, hKey, factor_input, num_channels_input*sizeof(float), MODE_READ, NULL);

  for (i=0 ; i<num_channels_output ; i++)
    factor_output[i] = 1.f;
  db_merge_data(hDB, hKeyRoot, "Settings/Output Factor", 
                factor_output, num_channels_output*sizeof(float), num_channels_output, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Settings/Output factor", &hKey);
  db_open_record(hDB, hKey, factor_output, num_channels_output*sizeof(float), MODE_READ, NULL);

  /*---- Create/Read variables ----*/

  /* Input */
  db_merge_data(hDB, hKeyRoot, "Variables/Input", 
                var_input, num_channels_input*sizeof(float), num_channels_input, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Variables/Input", &hKeyInput);
  memcpy(input_mirror, var_input, num_channels_input*sizeof(float));

  /* Output */
  db_merge_data(hDB, hKeyRoot, "Variables/Output", 
                var_output, num_channels_output*sizeof(float), num_channels_output, TID_FLOAT);
  db_find_key(hDB, hKeyRoot, "Variables/Output", &hKeyOutput);

  /*---- Initialize device drivers ----*/

  /* call init method */
  for (i=0 ; pequipment->driver[i].name[0] ; i++)
    {
    sprintf(str, "Settings/Devices/%s", pequipment->driver[i].name);
    status = db_find_key(hDB, hKeyRoot, str, &hKey);
    if (status != DB_SUCCESS)
      {
      db_create_key(hDB, hKeyRoot, str, TID_KEY);
      status = db_find_key(hDB, hKeyRoot, str, &hKey);
      if (status != DB_SUCCESS)
        {
        cm_msg(MERROR, "multi_init", "Cannot create %s entry in online database", str);
        free_mem();
        return FE_ERR_ODB;
        }
      }

    status = pequipment->driver[i].func(CMD_INIT, hKey, pequipment->driver[i].index, 
                                        pequipment->driver[i].channels);
    if (status != FE_SUCCESS)
      {
      free_mem();
      return status;
      }
    }

  /* compose device driver channel assignment */
  for (i=0,j=0,index=0,ch_offset=0 ; i<num_channels_input ; i++,j++)
    {
    while (pequipment->driver[index].name[0] &&
           (j >= pequipment->driver[index].channels ||
           pequipment->driver[index].type != CH_INPUT))
      {
      ch_offset += j;
      index++;
      j = 0;
      }

    driver_input[i] = pequipment->driver[index].func;
    device_index_input[i] = pequipment->driver[index].index;
    channel_offset_input[i] = ch_offset;
    }
  for (i=0,j=0,index=0,ch_offset=0 ; i<num_channels_output ; i++,j++)
    {
    while (pequipment->driver[index].name[0] &&
           (j >= pequipment->driver[index].channels ||
           pequipment->driver[index].type != CH_OUTPUT))
      {
      ch_offset += j;
      index++;
      j = 0;
      }

    driver_output[i] = pequipment->driver[index].func;
    device_index_output[i] = pequipment->driver[index].index;
    channel_offset_output[i] = ch_offset;
    }

  /* set initial demand values */
  for (i=0 ; i<num_channels_output ; i++)
    DRIVER_OUTPUT(i)(CMD_SET, device_index_output[i], i-channel_offset_output[i], var_output[i]);

  /* open hot link to output record */
  db_open_record(hDB, hKeyOutput, var_output, num_channels_output*sizeof(float), 
                 MODE_READ, multi_output);

  /* initially read all channels */
  multi_read(-1);

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT multi_exit()
{
INT i;

  free_mem();

  /* call exit method of device drivers */
  for (i=0 ; pequipment->driver[i].func != NULL ; i++)
    pequipment->driver[i].func(CMD_EXIT, pequipment->driver[i].index);

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT multi_idle()
{
static int channel=0;

  channel = (channel + 1) % num_channels_input;
  multi_read(channel);

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT cd_multi_read(char *pevent)
{
  memcpy(pevent, var_input, sizeof(float)*num_channels_input);
  pevent += sizeof(float)*num_channels_input;
  
  memcpy(pevent, var_output, sizeof(float)*num_channels_output);
  pevent += sizeof(float)*num_channels_output;

  return sizeof(float)*(num_channels_input+num_channels_output);
}

/*----------------------------------------------------------------------------*/

INT cd_multi(INT cmd, PTYPE adr)
{
INT  status;

  switch (cmd)
    {
    case CMD_INIT:
      pequipment = (EQUIPMENT *) adr;
      status = multi_init();
      break;

    case CMD_EXIT:
      status = multi_exit();
      break;


    case CMD_IDLE:
      status = multi_idle();
      break;

    default:
      cm_msg(MERROR, "Multimeter class driver", "Received unknown command %d", cmd);
      status = FE_ERR_DRIVER;
      break;

    }

  return status; 
}
