/********************************************************************\

  Name:         hyt1331.h
  Created by:   Stefan Ritt

  Cotents:      Declarations for CAMAC functions for HYTEC 1331
                controller.

  Revision history
  ------------------------------------------------------------------
  date        by    modification
  --------    ---   ------------------------------------------------
  22-DEC-93   SR    created
  16-JUL-95   SR    fixed bug in CAMO(..., data[i]) where _DX = CAMAC_BASE
                    was overwritten by index calculation
  16-JUL-95   SR    fixed bug in CAMI(..., &data[i]) where _AX con-
                    taining data was overwritten by index calculation
  09-JUL-96   SR    Added routines for multi crate access
  14-NOV-96   SR    Added {} brackets around all macros to work
                    poperly in for and if statements
  27-NOV-96   SR    Added a local _data variable for all 24bit operations
                    due to problems with cami24(..., pdata++)

\********************************************************************/

/**** General Data types ********************************************/
#ifndef BYTE_DEFINED
#define BYTE_DEFINED
typedef unsigned char      BYTE;
typedef          int       BOOL;
typedef unsigned short int WORD;
typedef unsigned long int  DWORD;
#define TRUE  1
#define FALSE 0
#endif

/*------------------------------------------------------------------*/

/*
  Base address of PC card. Must match jumper setting:
  ===================================================

  LK1    LK2    ADDRESS

  IN     IN     0x200 - 0x23F
  OUT    IN     0x280 - 0x3BF
  IN     OUT    0x300 - 0x33F
  OUT    OUT    0x380 - 0x3BF
*/

#define CAMAC_BASE 0x300

/*
  Interrupt request selection. Must match jumper setting:
  =======================================================

  LINK IN       IRQ line    Software interrupt

   9            IRQ 5       INT 0x0D (13)
  10            IRQ 3       INT 0x0B (11)
  11            IRQ 9       INT 0x02 (Cascaded)
  12            IRQ 12      INT 0x02 (Cascaded)
*/

#define LAM_INT 0x0D

/*------------------------------------------------------------------*/

#ifndef CAMAC_BASE_DEFINED
#define CAMAC_BASE_DEFINED
WORD _cam_base = CAMAC_BASE;
#else
extern WORD _cam_base;
#endif

/*------------------------------------------------------------------*/

/*
   macro                purpose

   output               address  data     status
   ======               =======  ======   ========

   CAMO8              : n, a, f   8 bit   -
   CAMO               : n, a, f  16 bit   -
   CAMO24             : n, a, f  24 bit   -

   CAMOF8             : f         8 bit   -
   CAMOF              : f        16 bit   -
   CAMOF24            : f        24 bit   -

   CAMOQ8             : n, a, f   8 bit   X, Q
   CAMOQ              : n, a, f  16 bit   X, Q
   CAMOQ24            : n, a, f  24 bit   X, Q

   CAMO_RSTART        : n, a, f  16 bit   -     Start repeat cycle
   CAMO_REPEAT        : -        16 bit   -     Repeat cycle

   input
   =====

   CAMI8              : n, a, f   8 bit   -
   CAMI               : n, a, f  16 bit   -
   CAMI24             : n, a, f  24 bit   -

   CAMIF8             : f         8 bit   -
   CAMIF              : f        16 bit   -
   CAMIF24            : f        24 bit   -

   CAMIQ8             : n, a, f   8 bit   X, Q
   CAMIQ              : n, a, f  16 bit   X, Q
   CAMIQ24            : n, a, f  24 bit   X, Q

   CAMI_RSTART        : n, a, f  16 bit   -     Start repeat cycle
   CAMI_REPEAT        : -        16 bit   -     Repeat cycle
   CAMI_LOOP          : n, a1, a2, f  16 bit    Loop subaddress

   misc
   ====

   CAMAC_N            : n                 set n for following operations
   CAMAC_A            : a                 set a for following operations
   CAMAC_F            : f                 set f for following operations
   CAMAC_S            :                   read back status in AX

   CAMAC_SET_CRATE                        set camac crate 0..3

   CAMAC_SET_INHIBIT
   CAMAC_CLEAR_INHIBIT
   CAMAC_C_CYCLE
   CAMAC_Z_CYCLE
   CAMAC_IRQ_ENABLE
   CAMAC_IRQ_DISABLE
   CAMAC_SET_IRQ_BIT(station)             set mask bit for station
   CAMAC_CLEAR_IRQ_BIT(station)           clear mask bit for station
   CAMAC_READ_LAM()                       read encoded lam

*/

/*------------------------------------------------------------------*/

#define CAMAC_N(_n)  \
  _DX = _cam_base+8; \
  _AL = _n;          \
  __emit__(0xEE);

#define CAMAC_A(_a)  \
  _DX = _cam_base+6; \
  _AL = _a;          \
  __emit__(0xEE);

#define CAMAC_F(_f)  \
  _DX = _cam_base+10;\
  _AL = _f;          \
  __emit__(0xEE);

#define CAMAC_S      \
  _DX = _cam_base+6; \
  __emit__(0xEC);

/*------------------------------------------------------------------*/

#define CAMO8(_n, _a, _f, _d) \
{ _AL = (BYTE) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMO(_n, _a, _f, _d) \
{ _AX = (WORD) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMO24(_n, _a, _f, _d) \
{ DWORD _data;       \
  _data = _d;        \
  _AX = (WORD) (_data); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  _AL = (BYTE) (_data>>16); \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

/*------------------------------------------------------------------*/

#define CAMOF8(_f, _d) \
{ _AL = (BYTE) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMOF(_f, _d) \
{ _AX = (WORD) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMOF24(_f, _d) \
{ DWORD _data;       \
  _data = _d;        \
  _AX = (WORD) (_data); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  _AL = (BYTE) (_data>>16); \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEE);    \
  CAMAC_F(_f);       \
  CAMAC_S; }

/*------------------------------------------------------------------*/

#define CAMOQ8(_n, _a, _f, _d, _x, _q) \
{ _AL = (BYTE) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_q) = _AL;       \
  *(_q) &= 1;        \
  __emit__(0xB1,7);  \
  __emit__(0xD3,0xF8);\
  *(_x) = _AL; }

#define CAMOQ(_n, _a, _f, _d, _x, _q) \
{ _AX = (WORD) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_q) = _AL;       \
  *(_q) &= 1;        \
  __emit__(0xB1,7);  \
  __emit__(0xD3,0xF8);\
  *(_x) = _AL; }

#define CAMOQ24(_n, _a, _f, _d, _x, _q) \
{ DWORD _data;       \
  _data = _d;        \
  _AX = (WORD) (_data); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  _AL = (BYTE) (_data>>16); \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_q) = _AL;       \
  *(_q) &= 1;        \
  __emit__(0xB1,7);  \
  __emit__(0xD3,0xF8);\
  *(_x) = _AL; }

/*------------------------------------------------------------------*/

#define CAMI8(_n, _a, _f, _d) \
{ unsigned char _tmp;\
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0xEC);    \
  _tmp = _AL;        \
  *(_d) = _tmp; }

#define CAMI(_n, _a, _f, _d) \
{ unsigned _tmp;     \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _tmp = _AX;        \
  *(_d) = _tmp; }

#define CAMI24(_n, _a, _f, _d) \
{ DWORD _data;       \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _data = (DWORD) _AX; \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  *(((WORD *)&_data)+1) = (unsigned char) _AX; \
  *_d = _data; }

/*------------------------------------------------------------------*/

#define CAMIF8(_f, _d) \
{ unsigned char _tmp;\
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0xEC);    \
  _tmp = _AL;        \
  *(_d) = _tmp; }

#define CAMIF(_f, _d) \
{ unsigned _tmp;     \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _tmp = _AX;        \
  *(_d) = _tmp; }

#define CAMIF24(_f, _d) \
{ DWORD _data;       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _data = (DWORD _AX; \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  *(((WORD *)&_data)+1) = (unsigned char) _AX; \
  *_d = _data; }

/*------------------------------------------------------------------*/

#define CAMIQ8(_n, _a, _f, _d, _x, _q) \
{ unsigned char _tmp;\
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _BL = _AL;         \
  _DX = _cam_base;   \
  __emit__(0xEC);    \
  _tmp = _AL;        \
  *(_d) = _tmp;      \
  _AL = _BL;         \
  *(_q) = _AL;       \
  *(_q) &= 1;        \
  __emit__(0xB1,7);  \
  __emit__(0xD3,0xF8);\
  *(_x) = _AL; }

#define CAMIQ(_n, _a, _f, _d, _x, _q) \
{ unsigned _tmp;     \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _BL = _AL;         \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _tmp = _AX;        \
  *(_d) = _tmp;      \
  _AL = _BL;         \
  *(_q) = _AL;       \
  *(_q) &= 1;        \
  __emit__(0xB1,7);  \
  __emit__(0xD3,0xF8);\
  *(_x) = _AL; }

#define CAMIQ24(_n, _a, _f, _d, _x, _q) \
{ DWORD _data;       \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _BL = _AL;         \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _data = (DWORD _AX;\
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  *(((WORD *)&_data)+1) = (unsigned char ) _AX; \
  *_d = _data;       \
  _AL = _BL;         \
  *(_q) = _AL;       \
  *(_q) &= 1;        \
  __emit__(0xB1,7);  \
  __emit__(0xD3,0xF8);\
  *(_x) = _AL; }

/*------------------------------------------------------------------*/

#define CAMI_RSTART(_n, _a, _f, _d) \
{ unsigned _tmp;     \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _DX = _cam_base;   \
  __emit__(0x42);    \
  __emit__(0x42);    \
  __emit__(0xEC);    \
  _AH = _AL;         \
  __emit__(0x4A);    \
  __emit__(0x4A);    \
  __emit__(0xEC);    \
  _tmp = _AX;        \
  *(_d) = _tmp;      \
  _DX = _cam_base+12;\
  __emit__(0xED); }

#define CAMI_REPEAT(_d) \
  { unsigned _tmp;   \
  _DX = _cam_base+12;\
  __emit__(0xED);    \
  _tmp = _AX;        \
  *(_d) = _tmp; }

/*------------------------------------------------------------------*/

#define CAMIQ_RSTART(_n, _a, _f) \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;

#define CAMIQ_REPEAT(_q) \
  { unsigned _tmp;   \
  _DX = _cam_base+12;\
  __emit__(0xED);    \
  CAMAC_S;           \
  _tmp = _AL & 1;    \
  *(_q) = _tmp; }

/*------------------------------------------------------------------*/

#define CAMO_RSTART(_n, _a, _f, _d) \
{ _AX = (WORD) (_d); \
  _DX = _cam_base;   \
  __emit__(0xEE);    \
  __emit__(0x42);    \
  __emit__(0x42);    \
  _AL = _AH;         \
  __emit__(0xEE);    \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMO_REPEAT(_d) \
  { _DX = _cam_base+12;\
  _AX = _d;          \
  __emit__(0xEF); }

/*------------------------------------------------------------------*/

/*
  CAMI_LOOP reads out consecutive subaddresses in the range a1..a2.
  Note that using the 1331 repeat command on different addresses
  is tricky since the 1331 doesn't accept a subaddress change if
  a previous repeat cycle has not yet completed. Therefore, a
  outport command is issued after the CAMI_RSTART to a dummy
  address to make a delay. The subaddress changes in the loop are
  delayed by the pdata++ command, whichs takes some time. This may
  not be true on faster computers (>66 MHz 486).
  Note also that the CAMI_REPEAT command reads in the result of
  the PREVIOUS repeat cycle and starts the a new one. So all data
  has to be shifted by one.
*/

#define CAMI_LOOP(_n, _a1, _a2, _f, _d) \
  {                                     \
  WORD a, *pdata;                       \
                                        \
  pdata = (_d);                         \
  CAMI_RSTART(_n, _a1, _f, pdata);      \
                                        \
  _DX = _cam_base+15;                   \
  __emit__(0xEE);                       \
                                        \
  CAMAC_A(_a1+1);                       \
  CAMI_REPEAT(pdata++);                 \
                                        \
  for (a=_a1+2 ; a<=_a2+1 ; a++)        \
    {                                   \
    CAMAC_A(a);                         \
    CAMI_REPEAT(pdata++);               \
    }                                   \
  }

/*------------------------------------------------------------------*/

#define CAMAC_SET_CRATE(_crate) \
  _cam_base = CAMAC_BASE + 0x10*_crate; \

#define CAMAC_SET_INHIBIT()   \
{ _DX = _cam_base+10;         \
  _AL = 33;                   \
  __emit__(0xEE); }

#define CAMAC_CLEAR_INHIBIT() \
{ _DX = _cam_base+10;         \
  _AL = 32;                   \
  __emit__(0xEE); }

#define CAMAC_C_CYCLE()       \
{ _DX = _cam_base+10;         \
  _AL = 36;                   \
  __emit__(0xEE); }

#define CAMAC_Z_CYCLE()       \
{ _DX = _cam_base+10;         \
  _AL = 34;                   \
  __emit__(0xEE); }

#define CAMAC_IRQ_ENABLE()    \
{ _DX = _cam_base+10;         \
  _AL = 41;                   \
  __emit__(0xEE); }

#define CAMAC_IRQ_DISABLE()   \
{ _DX = _cam_base+10;         \
  _AL = 40;                   \
  __emit__(0xEE); }

#define CAMAC_SET_IRQ_BIT(station) \
{ _DX = _cam_base+10;         \
  _AL = 64+station;           \
  __emit__(0xEE); }

#define CAMAC_CLEAR_IRQ_BIT(station) \
{ _DX = _cam_base+10;         \
  _AL = 128+station;          \
  __emit__(0xEE); }

#define CAMAC_READ_LAM()      \
{ _DX = _cam_base+8;          \
  __emit__(0xEC); }
