/*********************************************************************

  Name:         lrs1821.c
  Created by:   Stefan Ritt

  Cotents:      Routines for accessing Fastbus over LeCroy 1821 SMI
                

  Revision history
  ------------------------------------------------------------------
  date        by    modification
  --------    ---   ------------------------------------------------
  18-MAY-93   SR    created
  09-JUL-96   SR    added ECL in and out routines, simplified patch
                    table structure

*********************************************************************/

#pragma inline

#include <ctype.h>
#include <stdio.h>
#include <dos.h>

/* midas include file */
#include "midas.h"
#define BYTE_DEFINED
#include "lrs1821.h"

static struct {
  WORD  addr;
  char  data[17];
} patch_table[] = {

  0x34, "c00400000708", // ECL OUT1=1
  0x35, "c00400000308", // ECL OUT1=0

  0x36, "800401020000", // load pdreg=01
  0x37, "830400000000", // test IN1
  0x38, "a3043a000000", // return if IN1 on
  0x39, "800400020000", // load pdreg=00
  0x3a, "c00400",

  0x40, "800000080000", // Hardware block transfer
  0x41, "80040000023f",
  0x42, "85b400020200",
  0x43, "a5b443040301",
  0x44, "807700000e20",
  0x45, "a5874a000f01",
  0x46, "a07745000a00",
  0x47, "850400000301",
  0x48, "d0047a00021f",
  0x49, "c00400000000",
  0x4a, "80041e020000",
  0x4b, "c58440000301",

  0x7a, "819400028609",

  0, "", // end marker
};

BOOL fb_init(void)
/*********************************************************************
* Initialize Fastbus including 1821 and 1691                         *
*********************************************************************/
{
WORD status, version, update, i, data, addr;
char *pc, str[4];

  // set 1821 front panel address
  outp(RCL, BYPASS + 0);

  // clear registers
  OUT(REG0, 0);
  OUT(REG1, 0);
  OUT(REG2, 0);
  OUT(REG3, 0);
  OUT(REG4, 0);
  OUT(REG5, 0);
  OUT(REG6, 0);
  OUT(REG7, 0);

  // Check communication
  OUT(REG3, _RESET);
  status = IN(REG3);
  if (status != _RESET)
    {
    cm_msg(MERROR, "fb_init", "Cannot access 1821. Check power and cabling.");
    return(FALSE);
    }

  // Check Fastbus power
  status = IN(REG7);
  if (!(status & PWR_1) || !(status & PWR_2))
    {
    cm_msg(MERROR, "fb_init", "Fastbus power test failed. Check power and cabling.");
    return(FALSE);
    }

  // Load ROM to sequencer
  OUT(REG0, PD_PMAR_TO_SEQ | PA_MENU_TO_SEQ | PLOAD + ROM_NO);
  OUT(REG7, PMADZ | PMADL);

  delay(1);
  while (IN(REG7) & ACTIVE);

  // check version of PROM
  OUT(REG0, PA_HOST_TO_SEQ | PD_MENU_TO_HOST + ROM_NO);

  OUT(REG1, ROM_VER); // Address of version #
  version = IN(REG2) & 0xFF;

  OUT(REG1, ROM_UPD); // Address of update #
  update = IN(REG2) & 0xFF;

  if (version != 6 && version != 7)
    fprintf(stderr, "fb_init: Unkown PROM version %d.%d.\n", version, update);

  // load patch table
  OUT(REG0, PA_HOST_TO_SEQ | PD_HOST_TO_SEQ | PLOAD);
  for (i=0 ; patch_table[i].addr > 0; i++)
    {
    pc = patch_table[i].data;
    addr = patch_table[i].addr*8;
    do
      {
      if (isxdigit(*pc) && isxdigit(*(pc+1)))
        {
        str[0] = *pc;
        str[1] = *(pc+1);
        str[2] = 0;
        sscanf(str, "%X", &data);

        OUT(REG1, addr);
        OUT(REG2, data);
        OUT(REG7, SQWRT);
        }
      pc+=2;
      addr++;
      } while (*pc);
    }

  // ignite the sequencer
  OUT(REG0, PA_HOST_TO_SEQ | PD_HOST_TO_SEQ | PLOAD);
  OUT(REG1, IDLE);
  OUT(REG0, PA_HOST_TO_SEQ | PD_HOST_TO_SEQ);

  // clear all fastbus lines
  fb_exec(FPCLBS);

  // set arbitration level
  OUT(REG2, ARB_LEVEL);
  fb_exec(FBARB);

  // reset bus
  fb_exec(FPRBUS);

  return(TRUE);
}


void fb_write_data_memory(WORD *buffer, WORD count)
/*********************************************************************
* Write 'count' words to 1821 data memory from 'buffer'              *
*********************************************************************/
{
int           i, j;

  OUT(REG3, _RESET | DMB_HOST_TO_DMEM);
  OUT(REG6, EAUTO | UP + 0);
  OUT(REG7, DMARL);

  for (i=0,j=0 ; i<count/2 ; i++)
    {
    OUT(REG4, buffer[j]);
    j++;
    OUT(REG5, buffer[j]);
    j++;

    OUT(REG7, DMWRT);
    }
}

void fb_read_data_memory(WORD *buffer, WORD count)
/*********************************************************************
* Read 'count' words to 1821 data memory from 'buffer'               *
*********************************************************************/
{
#undef READ_VIA_DMA

#ifdef READ_VIA_DMA // ----------------- DMA Reaout ------------------

WORD     page, ofs, seg;
DWORD    adr;
WORD far *p;

  if (count == 0)
    return;

  OUT(REG3, _RESET | MUXR0 | DMB_DMEM_TO_HOST);
  OUT(REG6, EAUTO | UP | EGIP5 + 0);
  OUT(REG7, DMARL);

  p = buffer;
  seg = FP_SEG(p);
  ofs = FP_OFF(p);

  adr = seg * 16L + ofs;
  page = *((char *) (&adr) + 2);
  count = count * 2 - 1;           // word -> byte minus one for DMA

  // check if buffer on boundary crossing
  if ((adr & 0xFFFF) + count > 0xFFFF)
    {
    fprintf(stderr, "fb_read_data_memory: buffer on DMA boundary crossing\n");
    return;
    }

  outp(0x00C, 0);                  // reset f/f
  outp(0x00B, 0x05);               // write channel 1 in demand mode

  outp(0x002, (adr & 0xFF));       // set address (offset, page)
  outp(0x002, (adr >> 8));
  outp(0x083, page);

  outp(0x003, (count & 0xFF));     // set count
  outp(0x003, (count >> 8));

  outp(0x00A, 1);                  // enable DMA

  outp(DMA, 0);                    // initiat DMA transfer for 1691
  while ((inp(0x008) & 2) == 0);   // wait until completed


#else //------------------------ IO Readout --------------------------

WORD far *p;

  if (count == 0)
    return;

  OUT(REG3, _RESET | MUXR0 | DMB_DMEM_TO_HOST);
  OUT(REG6, EAUTO | UP | EGIP5 + 0);
  OUT(REG7, DMARL);

  p = buffer;

  asm push di
  asm les  di,p
  asm mov  cx,count
  asm mov  dx,REG4
  asm cld
loop1:

  asm in   ax,dx
  asm stosw

  asm dec  cx
  asm jnz  loop1:
  asm pop  di

#endif
}

BOOL fb_exec(WORD location)
/*********************************************************************
* Execute sequencer function at "location" in 1821                   *
*********************************************************************/
{
int i;

  OUT(REG1, location*8);
  OUT(REG7, GO);

  // check if sequencer executed start, ~100 us timeout
  for (i=100 ; (IN(REG7) & IWARDY) && i>0 ; i--);
  if (i==0)
    {
    fprintf(stderr, "fb_exec: sequencer timeout.\n");
    return(FALSE);
    }
  return(TRUE);
}

BOOL fb_load32(DWORD data)
/*********************************************************************
* Load a 32-bit data word to the sequencer. Used in read/write fast- *
* bus cycles.                                                        *
*********************************************************************/
{
  OUT(REG2, (WORD) data);
  if (!fb_exec(LD_LOW))
    return(FALSE);

  OUT(REG2, (WORD) (data>>16));
  if (!fb_exec(LD_HIGH))
    return(FALSE);

  return(TRUE);

}

BOOL fb_fpad(WORD slot)
/*********************************************************************
* Primary address cycle to data space                                *
*********************************************************************/
{
  OUT(REG2, slot);
  OUT(REG1, FPAD*8);
  OUT(REG7, GO);

  return(TRUE);
}

BOOL fb_fpac(WORD slot)
/*********************************************************************
* Primary address cycle to control space                             *
*********************************************************************/
{
  OUT(REG2, slot);
  OUT(REG1, FPAC*8);
  OUT(REG7, GO);

  return(TRUE);
}

BOOL fb_fpwnta(DWORD adr)
/*********************************************************************
* Seconadry address write cycle                                      *
*********************************************************************/
{
  fb_load32(adr);
  return(fb_exec(FPWNTA));
}

BOOL fb_fprnta(DWORD *adr)
/*********************************************************************
* Read secondary address from NTA (Next Transfer Address) register   *
*********************************************************************/
{
  // no pedestal subtraction, reg. 4 & 5 dest of ped.subr.subsyst. output
  OUT(REG3, _RESET | MWENB | DMB_PED_TO_HOST);

  // no auto advance due to read
  OUT(REG6, 0);    

  // data memory addr load from reg. 6
  OUT(REG7, DMARL);

  if (!fb_exec(FPRNTA))
    return(FALSE);

  ((WORD *)adr)[0] = IN(REG4);
  ((WORD *)adr)[1] = IN(REG5);

  return(TRUE);
}


BOOL fb_fpw(DWORD data)
/*********************************************************************
* Execute a single random data write cycle. It is assumed that the   *
* module of interest has already been addressed correctly.           *
*********************************************************************/
{
  fb_load32(data);

  return(fb_exec(FPW));
}

BOOL fb_fpr(DWORD *data)
/*********************************************************************
* Execute a single random data read cycle. It is assumed that the    *
* module of interest has already been addressed correctly.           *
*********************************************************************/
{
  // no pedestal subtraction, reg. 4 & 5 dest of ped.subr.subsyst. output
  OUT(REG3, _RESET | MWENB | DMB_PED_TO_HOST);

  // no auto advance due to read
  OUT(REG6, 0);

  // data memory addr load from reg. 6
  OUT(REG7, DMARL);

  if (!fb_exec(FPR))
    return(FALSE);

  ((WORD *)data)[0] = IN(REG4);
  ((WORD *)data)[1] = IN(REG5);

  return(TRUE);
}

BOOL fb_fprbl(WORD *data, WORD *count, WORD flags)
/*********************************************************************
* Execute a FASTBUS block read into 1821 internal memory. It is      *
* assumed that the module of interest has already been addressed     *
* correctly.                                                         *
*********************************************************************/
{
int i;

  // Configure 1821
  OUT(REG3, _RESET | MWENB | DMB_PED_TO_DMEM | flags);
  OUT(REG6, EAUTO | UP + 0);  // start at zero
  OUT(REG7, DMARL);

//  OUT(REG1, USER);
//  OUT(REG1, HDRBL);
  OUT(REG1, FPRBL*8);     // FPRBL for non-LeCroy modules (slower)
  OUT(REG7, GO);

  // wait for 1821 to complete, ~200 us timeout
  for (i=200 ; (IN(REG7) & (IWARDY | ACTIVE)) && i>0 ; i--);

  if (i==0)
    {
    fprintf(stderr, "fb_fprbl: block transfer timeout.\n");
    return(FALSE);
    }

  // read status
  OUT(REG0, PA_HOST_TO_SEQ | PD_SEQ_TO_HOST);
  if (IN(REG2) != 0x200)
    {
    fprintf(stderr, "fb_fprbl: bad block transfer status.\n");
    return(FALSE);
    }
  OUT(REG0, PA_HOST_TO_SEQ | PD_HOST_TO_SEQ);

  // read word count
  i = (IN(REG6) & 0xFFF) * 2;
  if (i > *count)
    i = *count;
  *count = i;

  fb_read_data_memory(data, *count);

  return(TRUE);
}

BOOL fb_ped_load(WORD *data, WORD count, WORD slot, BOOL high)
/*********************************************************************
* Download 'count' pedestal values in 'data' on 'slot'               *
* If 'high'==TRUE, put to the high range pedestal memory.            *
*********************************************************************/
{
int  i;
WORD addr;

  OUT(REG3, _RESET | PDMSEL | DMB_HOST_TO_PED);

  for (i=0 ; i<count ; i++)
    {
    addr = (slot << 11) | (high << 7) | i;
    OUT(REG5, addr);
    OUT(REG4, data[i]);
    OUT(REG7, PDMADR);
    OUT(REG7, PDMWRT);
    }

  // verify data
  for (i=0 ; i<count ; i++)
    {
    OUT(REG3, _RESET | PDMSEL | DMB_HOST_TO_PED);
    addr = (slot << 11) | (high << 7) | i;
    OUT(REG5, addr);
    OUT(REG7, PDMADR);

    OUT(REG3, _RESET | PDMSEL | DMB_PED_TO_HOST);
    if (IN(REG4) != data[i])
      {
      printf("fb_ped_load: pedestal verify error.\n");
      return(FALSE);
      }
    }

  return(TRUE);
}

BOOL fb_thr_load(WORD data, BOOL high)
/*********************************************************************
* Download threshold value                                           *
*********************************************************************/
{
  OUT(REG3, _RESET | DMB_HOST_TO_PED);
  OUT(REG5, (high << 7));
  OUT(REG4, data);
  OUT(REG7, PDMADR);
  OUT(REG7, PDCWRT);
  return(TRUE);
}

BOOL fb_in1()
/*********************************************************************
* Read ECL IN1                                                       *
*********************************************************************/
{
int s;

  fb_exec(IN1TST);

  // read status
  OUT(REG0, PA_HOST_TO_SEQ | PD_SEQ_TO_HOST);
  s = IN(REG2);
  OUT(REG0, PA_HOST_TO_SEQ | PD_HOST_TO_SEQ);

  return(s);
}
/*------------------------ time functions --------------------------*/

void delay_us(WORD delay_time)
/*********************************************************************
* delay in units of 1us. Minimal 100, Maximal 50000 (=50ms)          *
*********************************************************************/
{
  if ((delay_time>=100) && (delay_time<=50000U))
    {
    delay_time=delay_time/0.8381+1000;
    disable();
    outp(0x43,0x34);
    outp(0x40,delay_time & 255);
    outp(0x40,delay_time >> 8);
    outp(0x43,4);
    while ((unsigned)inp(0x40)+(unsigned)(inp(0x40) << 8) > 1000);
    enable();
    }
}

