/********************************************************************\

  Name:         lrs4032.c
  Created by:   Stefan Ritt

  Contents:     LeCroy LRS 4032 High Voltage Device Driver


  Revision history
  -------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  26-MAR-97    SR    created


\********************************************************************/

#include <stdio.h>
#include <math.h>
#include "midas.h"
#include "msystem.h"
#include "y_sc_user.h"
#if __NORPC__
 #include "cam8210.h"
#endif
#define FRAME_TO_CHANNEL 0
#define CHANNEL_TO_FRAME 1

#define LRS4032_STR "\
BCN_Mainframe = STRING : [32] 0 7 1 63\n\
"
/*---- globals -----------------------------------------------------*/
void string_parse(char *line, int *ntoken, char *p,char *token[]);

/* Specific LRS4032 HV control structure */
struct {
  INT    mainframe[16];
  INT    ext_hv;
  INT    branch;
  INT    crate;
  INT    station;
} lrs;

/*---- device driver routines ------------------------------------------------*/

INT lrs4032_init(HNDLE hKeyRoot, INT channels)
{
  char chain_address[256];
  int   status, size, i, q, n_cards;
  char  str[256];
  HNDLE hDB;
  
  cm_get_experiment_database(&hDB, NULL);
  
  /* Get Mainframe address */
  {
    INT ntoken, i, ii;
    char * token[16], *p;

    status = db_create_record(hDB, hKeyRoot, "", LRS4032_STR);
    if (status != DB_SUCCESS)
      {
	cm_msg(MERROR,"lrs4032_init","Cannot create BCN_Mainframe\n");
	return FE_ERR_ODB;
      }
    size = sizeof(chain_address);
    status = db_get_value(hDB, hKeyRoot, "BCN_Mainframe", chain_address, &size, TID_STRING);
    if (status != DB_SUCCESS)
      {
	cm_msg(MINFO, "lrs4032_init", "Cannot find 'BCN_Mainframe ' entry in ../Devices/LRS4032 database");
	return FE_ERR_ODB;
      }
    p = (char *) calloc(256,sizeof(char));
    string_parse(chain_address, &ntoken,  p, token);
    lrs.branch = atoi(token[0]);
    lrs.crate  = atoi(token[1]);
    lrs.station= atoi(token[2]);
    printf("LRS BCN_Mainframe %s - %ld %ld %ld - \n",chain_address, lrs.branch,lrs.crate,lrs.station);
    if (ntoken < 4)
      {
	printf(" Mainframe not defined:%s\n",chain_address);
	cm_msg(MINFO,  "lrs4032_init","Mainframe not defined in database");
	return FE_ERR_ODB;
      }    
    for (i=3;i<ntoken;i++)
      {
	lrs.mainframe[i-3] = atoi(token[i]);
	printf("Valid LRS4032 Mainframe address:%i\n",lrs.mainframe[i-3]);
      }
    free (p);
  }

  /* read LRS1440 settings */
  /* initialize RS232 port */
  /* check if module is living  */
  /* check if HV enabled */
  /* read rest of status */
  /* turn on HV main switch */
  
  /* save ext for HV slot */
#if __NORPC__
  CAM_BC (&lrs.ext_hv,lrs.branch,lrs.crate);

  /* Init LRS4032 */
  CAM_CMD(&lrs.ext_hv,lrs.station,0,24);        /* disable LAM_1 */
  CAM_CMD(&lrs.ext_hv,lrs.station,1,24);        /* disable LAM_1 */
  CAM_CMD(&lrs.ext_hv,lrs.station,0, 9);        /* Clear Buffer,L1,L2 */
#else
  caopen("vwchaos",&status);
  cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
  printf("Status:%d -E%xB%dC%dN%d\n",status,lrs.ext_hv,lrs.branch,lrs.crate,lrs.station);
  cfsa(24,lrs.ext_hv,&status,&q);
  cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,1);
  cfsa(24,lrs.ext_hv,&status,&q);
  cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
  cfsa(9,lrs.ext_hv,&status,&q);
  
#endif

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT lrs4032_exit()
{
  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------
  string_parse:
  return numberof token and tokens pointer in *p of the input line.
  ----------------------------------------------------------------------------*/

void string_parse(char *line, int *ntoken, char *p,char *token[])
{
  INT ntok=0,l;
  char tok[256];
  char *last_ptr, *tok_ptr;
  
  last_ptr = (char*) malloc(sizeof(char));
  tok_ptr  = tok;
  *last_ptr= ' ' ;
  l = strlen(line)+1;
  
  while (l--) {
    if ( (*line == ' ' ) || (*line == '\t') ||
	 (*line == '\n') || (*line == '\0' ) )
      {
	if ( (*last_ptr !=  ' ' &&  *last_ptr != '\t') ){
	  *tok_ptr = '\0';
	  strcpy(p,tok);
	  token[ntok++] = p;
	  p += (strlen(tok) +1);
	  tok_ptr = tok;
	}
      }
    else {
      *tok_ptr = *line ;
      tok_ptr++;
    }
    *last_ptr = *line++;
  }
  free(last_ptr);
  *ntoken = ntok;
  return;
}

/*----------------------------------------------------------------------------*/
INT lrs4032_assign_channel (INT direction, INT *channel, DWORD *frame, DWORD *chanframe)
{
  DWORD ch;
  
  ch = *channel;
  
  switch (direction)
    {
    case FRAME_TO_CHANNEL:
      printf("lrs4032_assign_channel: FRAME_TO_CHANNEL not implemented\n");
      break;
    case CHANNEL_TO_FRAME:
      *frame  = (ch / 32 );
      *chanframe = ch - (*frame)*32;
      if (*frame == 0)
	*frame = 0x10;
#ifdef DEBUG
      printf("%i ->",ch);
      printf(" %i %i ",*frame,*chanframe);
#endif
      break;
    }
  return 1;
}

/*----------------------------------------------------------------------------*/
INT lrs4032_set(INT channel, float fvalue)
{
  DWORD cmd, frame, fch, data;
  DWORD nlam, dummy, status, q, x;
  INT value;

  value = (INT)fvalue;

  status = 1;

  /* convert channel to frame and channel within frame */
  if (lrs4032_assign_channel(CHANNEL_TO_FRAME,&channel,&frame,&fch) < 0)
    return FE_ERR_ODB;

  /* compose data word for request */
  frame <<= 4;
  fch   <<= 10;

  /* set command */
  cmd = frame | fch | 0x0;
#ifdef DEBUG
  printf("WLRS4032:cmd:0x%x n:%i",cmd,lrs.station);
#endif
  nlam = q = 0;
  {
#if __NORPC__
    CAM16_OQX(&lrs.ext_hv,lrs.station,0,16,cmd,&q,&x);
#else
    cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
    cfsa(16,lrs.ext_hv,&cmd,&q);
#endif
    ss_sleep(100);
    if (q==0) status++;
    
    /* Write data */
    data = (value << 4) | 0x1;
#if __NORPC__
    CAM16_OQX(&lrs.ext_hv,lrs.station,0,16,data,&q,&x);
#else
    cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
    cfsa(16,lrs.ext_hv,&data,&q);
    ctstat(&x);
    x = (~x>>1)&1;
#endif

#ifdef DEBUG

    printf(" status:%i - value:0x%x - qx:%i%i ",status,data,q,x);
#endif
    ss_sleep(3000);
    status++;
    nlam++;
  }
  
  /* LAM check */
  q = nlam = 0;
  while (nlam < 100 && q != 1)
    {
#if __NORPC__
      CAM16_OQX(&lrs.ext_hv,lrs.station,1,27,dummy,&q,&x);
#else
    cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,1);
    cfsa(27,lrs.ext_hv,&dummy,&q);
    ctstat(&x);
    x = (~x>>1)&1;
#endif
      cm_yield(100); /* check for watchdog requests */
      status++;
      nlam++;
    }
#ifdef DEBUG
    printf("st:%i - nlam:%i - qx:%i%i\n",status,nlam,q,x);
#endif
  /* clear LAM */
#if __NORPC__
   CAM_CMD(&lrs.ext_hv,lrs.station,0,9);        /* Clear LAM_1 */
#else
    cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
    cfsa(9,lrs.ext_hv,&dummy,&q);
#endif

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT lrs4032_set_all(INT channels, float value)
{
  INT   i;
  
#ifdef DEBUG
  printf("in lrs4032_set_all\n");
#endif
  for (i=0 ; i<channels ; i++)
    {
      lrs4032_set(i, (INT) value);
      cm_yield(0); /* check for watchdog requests */
    }
  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT lrs4032_get(INT channel, float * pvalue)
{
  INT   value;
  float fvalue;
  DWORD cmd, frame, fch;
  DWORD nlam, dummy, status, q, x;
  WORD  dummi;

  status = 1;
  /* convert channel to frame and channel within frame */
  if (lrs4032_assign_channel(CHANNEL_TO_FRAME,&channel,&frame,&fch) < 0)
    return FE_ERR_ODB;

  /* compose data word for request */
  frame <<= 4;
  fch   <<= 10;

  /* set command */
  cmd = frame | fch | 0x3;
#ifdef DEBUG
  printf(" RLRS4032:cmd:%x ext:%x %i",cmd,lrs.ext_hv,lrs.station);
#endif

#if __NORPC__
  CAM16_OQX(&lrs.ext_hv,lrs.station,0,16,cmd,&q,&x);
#else
    cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
    cfsa(16,lrs.ext_hv,&cmd,&q);
#endif

  ss_sleep(500);
  if (q==0) status++;

  /* Set LAM check */
  nlam = q = 0;
  while (nlam < 5 && q != 1)
    {
#if __NORPC__
      CAM16_OQX(&lrs.ext_hv,lrs.station,1,27,dummy,&q,&x);
#else
      cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,1);
      cfsa(27,lrs.ext_hv,&dummy,&q);
#endif
      ss_sleep(100);
      status++;
      nlam++;
    }

  /* Read value */
  value = q = x = status = 0;
#if __NORPC__
  CAM16_IQX(&lrs.ext_hv,lrs.station,0,2,&dummi,&q,&x);
#else
  cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
  cfsa(2,lrs.ext_hv,&dummi,&q);
#endif
  value = dummi;
  ss_sleep(200);

  /* LAM check */
  q = nlam = 0;
  while (nlam < 10 && q != 1)
    {
#if __NORPC__
      CAM16_OQX(&lrs.ext_hv,lrs.station,1,27,dummy,&q,&x);
#else
      cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,1);
      cfsa(27,lrs.ext_hv,&dummy,&q);
      ctstat(&x);
      x = (~x>>1)&1;
#endif
      ss_sleep(50);
      status++;
      nlam++;
    }
  
  /* decode value */
  value >>= 4;
#ifdef DEBUG
  printf(" st:%i - value:%d - qx:%i%i\n",status,value,q,x);
#endif

  /* clear LAM */
#if __NORPC__
   CAM_CMD(&lrs.ext_hv,lrs.station,0,9);        /* Clear LAM_1 */
#else
    cdreg(&lrs.ext_hv,lrs.branch,lrs.crate,lrs.station,0);
    cfsa(9,lrs.ext_hv,&dummy,&q);
#endif

    *pvalue = (float) value;
  return FE_SUCCESS;
}

/*------------------------------------------------------------------*/

INT lrs4032_get_all(INT channels, float * array)
{
  float value;
  int  i;
  
#ifdef DEBUG
  printf("in lrs4032_get_all\n");
#endif

 /* read ALL this equip channels */
  for (i=0 ; i <= channels ; i++)
    {
      lrs4032_get(i, &value);
      array[i] = value;
      cm_yield(0); /* check for watchdog requests */
    }
  
  return FE_SUCCESS;
}

/*---- device driver entry point -----------------------------------*/

INT lrs4032(INT cmd, ...)
{
  va_list argptr;
  HNDLE   hKey;
  INT     channel, status;
  float   value, *pvalue;
  
  va_start(argptr, cmd);
  status = FE_SUCCESS;
  
  switch (cmd)
    {
    case CMD_INIT:
      hKey = va_arg(argptr, HNDLE);
      channel = va_arg(argptr, INT);
      status = lrs4032_init(hKey, channel);
      break;
      
    case CMD_EXIT:
      status = lrs4032_exit();
      break;
      
    case CMD_SET:
     channel = va_arg(argptr, INT);
      value   = (float) va_arg(argptr, double);
#ifdef DEBUG
      printf("in lrs4032 cmd Set -> :%i -chan:%i -Val:%f ",cmd,channel,value);
#endif
      status = lrs4032_set(channel, value);
     break;
      
    case CMD_SET_ALL:
      channel = va_arg(argptr, INT);
      value   = (float) va_arg(argptr, double);
#ifdef DEBUG
      printf("in lrs4032 cmd Set all -> :%i -chan:%i -Val:%f ",cmd,channel,value);
#endif
      status = lrs4032_set_all(channel, value);
      break;
      
    case CMD_GET:
      channel = va_arg(argptr, INT);
      pvalue  = va_arg(argptr, float*);
      status = lrs4032_get(channel, pvalue);
#ifdef DEBUG
      printf("in lrs4032 cmd Get <- :%i -Chan:%i -Val:%f ",cmd,channel,*pvalue);
#endif
      break;
      
    case CMD_GET_ALL:
      channel = va_arg(argptr, INT);
      pvalue  = va_arg(argptr, float*);
      status = lrs4032_get_all(channel, pvalue);
#ifdef DEBUG
      printf("in lrs4032 cmd get all <-:%i -Chan:%i -Val:%f ",cmd,channel,*pvalue);
#endif
      break;
      
      /* current measurements not supported in LRS4032 */
      
    case CMD_GET_CURRENT:
      break;
      
    case CMD_GET_CURRENT_ALL:
      break;
      
    case CMD_SET_CURRENT_LIMIT:
      break;
      
    default:
      cm_msg(MERROR, "lrs4032 device driver", "Received unknown command %d", cmd);
      status = FE_ERR_DRIVER;
      break;
    }
  
  va_end(argptr);
  
  return status;
}

/*------------------------------------------------------------------*/

INT ycd_hv_read(char *pevent)
{
  /*
  memcpy(pevent, demand, sizeof(float)*num_channels);
  pevent += sizeof(float)*num_channels;
  
  memcpy(pevent, measured, sizeof(float)*num_channels);
  pevent += sizeof(float)*num_channels;

  memcpy(pevent, current, sizeof(float)*num_channels);
  pevent += sizeof(float)*num_channels;
  */

  return (0);
}

/*------------------------------------------------------------------*/

